// *************************************************************************************************
//
// Horde3D
//   Next-Generation Graphics Engine
// --------------------------------------
// Copyright (C) 2006-2008 Nicolas Schulz
//
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//
// *************************************************************************************************

#include "egTextures.h"
#include "egModules.h"

#include "utDebug.h"


// *************************************************************************************************
// Class ImageLoader
// *************************************************************************************************

uint32 ImageLoader::nextSmallerPowerOfTwo( uint32 n )
{
	while( (n & (n - 1)) != 0 ) --n;

	return n;
}

#define MAKE_UINT(a,b,c,d) ((d<<24)|(c<<16)|(b<<8)|(a))

typedef struct  {
    uint32 dwSize;
    uint32 dwFlags;
    uint32 dwFourCC;
    uint32 dwRGBBitCount;
    uint32 dwRBitMask, dwGBitMask, dwBBitMask;
    uint32 dwRGBAlphaBitMask;
} DDPIXELFORMAT;

typedef struct  {
    uint32 dwCaps1;
    uint32 dwCaps2;
    uint32 Reserved[2];
} DDSCAPS2;

#define DDSD_MIPMAPCOUNT        0x00020000l

#define DDPF_RGB                0x00000040l
#define DDPF_FOURCC             0x00000004l
#define DDPF_ALPHAPIXELS        0x00000001l


typedef struct  {
    uint32 dwSize;
    uint32 dwFlags;
    uint32 dwHeight;
    uint32 dwWidth;
    uint32 dwPitchOrLinearSize;
    uint32 dwDepth;
    uint32 dwMipMapCount;
    uint32 dwReserved1[11];
    DDPIXELFORMAT ddpfPixelFormat;
    DDSCAPS2 ddsCaps;
    uint32 dwReserved2;
} DDSURFACEDESC2;

string ImageLoader::_failureReason;

uint32 ImageLoader::loadImage(const char* data, unsigned int size, void *&pixels, uint32& pixelsSize,
				int& width, int& height, TextureFormats::List& format, uint32& mipmaps, bool& cube)
{
	const uint32 headerSize = sizeof(uint32) + sizeof(DDSURFACEDESC2);
	if (size < headerSize)
	{
		_failureReason = "Buffer is smaller then header size";
		return 0;
	}
	uint32 dwMagic = *((uint32*)data);
	if (dwMagic!=MAKE_UINT('D','D','S',' '))
	{
		_failureReason = "Invalid DDS magic";
		return 0;
	}

	DDSURFACEDESC2 *desc = (DDSURFACEDESC2*)(data + sizeof(uint32));
	if ((desc->dwFlags & 0x1007)!=0x1007)
	{
		_failureReason = "Required fields missing";
		return 0;
	}
	if ((desc->dwFlags & 0x880008)!=0x0)
	{
		_failureReason = "Unsupported texture type";
		return 0;
	}
	
	width = desc->dwWidth;
	height = desc->dwHeight;
	mipmaps = (desc->dwFlags & DDSD_MIPMAPCOUNT) ? desc->dwMipMapCount : 1;

	if ((desc->ddpfPixelFormat.dwFlags == DDPF_RGB) &&
		(desc->ddpfPixelFormat.dwRGBBitCount == 32) &&
		(desc->ddpfPixelFormat.dwRBitMask == 0x00ff0000) &&
		(desc->ddpfPixelFormat.dwGBitMask == 0x0000ff00) &&
		(desc->ddpfPixelFormat.dwBBitMask == 0x000000ff))
		format = TextureFormats::BGRX8;
	else if ((desc->ddpfPixelFormat.dwFlags == (DDPF_RGB|DDPF_ALPHAPIXELS)) &&
		(desc->ddpfPixelFormat.dwRGBBitCount == 32) &&
		(desc->ddpfPixelFormat.dwRBitMask == 0x00ff0000) &&
		(desc->ddpfPixelFormat.dwGBitMask == 0x0000ff00) &&
		(desc->ddpfPixelFormat.dwBBitMask == 0x000000ff) &&
		(desc->ddpfPixelFormat.dwRGBAlphaBitMask == 0xff000000))
		format = TextureFormats::BGRA8;
	else if ((desc->ddpfPixelFormat.dwFlags == DDPF_RGB) &&
		(desc->ddpfPixelFormat.dwRGBBitCount == 32) &&
		(desc->ddpfPixelFormat.dwRBitMask == 0x000000ff) &&
		(desc->ddpfPixelFormat.dwGBitMask == 0x0000ff00) &&
		(desc->ddpfPixelFormat.dwBBitMask == 0x00ff0000))
		format = TextureFormats::RGBX8;
	else if ((desc->ddpfPixelFormat.dwFlags == (DDPF_RGB|DDPF_ALPHAPIXELS)) &&
		(desc->ddpfPixelFormat.dwRGBBitCount == 32) &&
		(desc->ddpfPixelFormat.dwRBitMask == 0x000000ff) &&
		(desc->ddpfPixelFormat.dwGBitMask == 0x0000ff00) &&
		(desc->ddpfPixelFormat.dwBBitMask == 0x00ff0000) &&
		(desc->ddpfPixelFormat.dwRGBAlphaBitMask == 0xff000000))
		format = TextureFormats::RGBA8;
	else if ( (desc->ddpfPixelFormat.dwFlags == DDPF_FOURCC) &&
		(desc->ddpfPixelFormat.dwFourCC == MAKE_UINT('D','X','T','1'))	)
		format = TextureFormats::DXT1;
	else if ( (desc->ddpfPixelFormat.dwFlags == DDPF_FOURCC) &&
		(desc->ddpfPixelFormat.dwFourCC == MAKE_UINT('D','X','T','3'))	)
		format = TextureFormats::DXT3;
	else if ( (desc->ddpfPixelFormat.dwFlags == DDPF_FOURCC) &&
		(desc->ddpfPixelFormat.dwFourCC == MAKE_UINT('D','X','T','5'))	)
		format = TextureFormats::DXT5;
	else if ( (desc->ddpfPixelFormat.dwFlags == DDPF_FOURCC) &&
		(desc->ddpfPixelFormat.dwFourCC == 113)	)
		format = TextureFormats::RGBA16F;
	else
	{
		_failureReason = "Unsupported texture format";
		return 0;
	}

	if (desc->ddsCaps.dwCaps2 == 0xfe00)
		cube = true;
	else if (desc->ddsCaps.dwCaps2 == 0)
		cube = false;
	else
	{
		_failureReason = "Unsupported texture type";
		return 0;
	}

	pixelsSize = size - headerSize;
	pixels = (void*)(data + headerSize);

	return 1;
}

string ImageLoader::getErrorString()
{
	return _failureReason;
}


// *************************************************************************************************
// Class Texture2DResource
// *************************************************************************************************

uint32 Texture2DResource::defTexObject = 0;


void Texture2DResource::initializationFunc()
{
	unsigned char defaultTexture[] = 
		{	255,192,128,0, 255,192,128,0, 255,192,128,0, 255,192,128,0,
			255,192,128,0, 255,192,128,0, 255,192,128,0, 255,192,128,0,
			255,192,128,0, 255,192,128,0, 255,192,128,0, 255,192,128,0,
			255,192,128,0, 255,192,128,0, 255,192,128,0, 255,192,128,0	};

	// Upload default texture
	defTexObject = Modules::renderer().uploadTexture2D(
		defaultTexture, sizeof(defaultTexture), 0, 4, 4, TextureFormats::RGBX8, false, true, true);
}


void Texture2DResource::releaseFunc()
{
	Modules::renderer().unloadTexture( defTexObject, false );
}


Texture2DResource::Texture2DResource( const string &name, int flags ) :
	Resource( ResourceTypes::Texture2D, name, flags )
{
	initDefault();
}


Texture2DResource::Texture2DResource( const string &name, int flags,
									  uint32 width, uint32 height, bool renderable ) :
	Resource( ResourceTypes::Texture2D, name, flags ), _rendBuf(0x0),
	_width(width), _height(height), _format(TextureFormats::RGBX8)
{	
	if( !renderable )
	{
		_texObject = Modules::renderer().uploadTexture2D( 0x0, 0, 1, _width, _height, _format,
			!(_flags & ResourceFlags::NoTexCompression), 
			!(_flags & ResourceFlags::NoTexFiltering), !(_flags & ResourceFlags::NoTexRepeat) );
		
		if( _texObject == 0 ) initDefault();
	}
	else
	{
		_loaded = true;
		_rendBuf = new RenderBuffer();
		*_rendBuf = Modules::renderer().createRenderBuffer(
			width, height, RenderBufferFormats::RGBA8, false, 1, !(_flags & ResourceFlags::NoTexFiltering), 0 ); 
		_texObject = _rendBuf->colBufs[0];
	}
}


Texture2DResource::~Texture2DResource()
{
	release();
}


void Texture2DResource::initDefault()
{
	_width = 0; _height = 0; _format = TextureFormats::RGBX8;
	_texObject = defTexObject; _rendBuf = 0x0;
}


void Texture2DResource::release()
{
	if( _rendBuf != 0x0 )
	{
		// In this case _texObject is just points to the render buffer
		Modules::renderer().destroyRenderBuffer( *_rendBuf );
		delete _rendBuf; _rendBuf = 0x0;
	}
	else if( _texObject != 0 && _texObject != defTexObject )
	{
		Modules::renderer().unloadTexture( _texObject, false );
	}

	_texObject = 0;
}


bool Texture2DResource::raiseError( const string &msg )
{
	// Reset
	release();
	initDefault();

	Modules::log().writeError( "Texture2D resource '%s': %s", _name.c_str(), msg.c_str() );
	
	return false;
}


bool Texture2DResource::load( const char *data, int size )
{
	if( !Resource::load( data, size ) ) return false;
	
	if( !Modules::config().loadTextures )
		return true;

//	bool makePOT = !Modules::renderer().supportsNPOTTextures();
//	if( _flags & ResourceFlags::NoTexPOTConversion ) makePOT = false;
	
	void *pixels;
	uint32 pixelsSize;
	bool cube;
	uint32 mipmaps;
	uint32 res = ImageLoader::loadImage( data, size, pixels, pixelsSize, _width, _height, _format, mipmaps, cube );
	if( res == 0 )
		return raiseError( "Invalid image format (" + ImageLoader::getErrorString() + ")" );
//	else if( res == 2 )
//		Modules::log().writeWarning( "Texture2D resource '%s': Texture size was changed to match POT", _name.c_str() );
	
	if (cube)
		return raiseError( "2D texture is expected" );

	if( !(_flags & ResourceFlags::NoTexMipmaps) && mipmaps == 1)
		mipmaps = 0;	//autogenerate mipmap chain

	// Upload texture
	_texObject = Modules::renderer().uploadTexture2D( pixels, pixelsSize, mipmaps, _width, _height, _format,
		!(_flags & ResourceFlags::NoTexCompression),
		!(_flags & ResourceFlags::NoTexFiltering), !(_flags & ResourceFlags::NoTexRepeat) );
	if( _texObject == 0 )
		return raiseError( "Failed to upload texture map" );

	return true;
}


bool Texture2DResource::updateData( int param, const void *data, int size )
{
	if( Resource::updateData( param, data, size ) ) return true;
	
	if( param == TextureResParams::PixelData )
	{
		//TODO: implement this
/*		if( _texObject == 0 ) return false;
		if( size != _width * _height * _comps ) return false;

		Modules::renderer().updateTexture2D( (unsigned char *)data, _width, _height, _comps,
			!(_flags & ResourceFlags::NoTexFiltering), _texObject );
*/
		return true;
	}

	return false;
}


float *Texture2DResource::downloadImageData()
{
	int width, height;
	return Modules::renderer().downloadTexture2DData( _texObject, &width, &height );
}


int Texture2DResource::getParami( int param )
{
	switch( param )
	{
	case TextureResParams::Width:
		return _width;
	case TextureResParams::Height:
		return _height;
	case TextureResParams::Format:
		return _format;
	default:
		return Resource::getParami( param );
	}
}


// *************************************************************************************************
// Class TextureCubeResource
// *************************************************************************************************

uint32 TextureCubeResource::defTexObject = 0;


void TextureCubeResource::initializationFunc()
{
	unsigned char defaultTexture[] = 
		{	255,192,128,0, 255,192,128,0, 255,192,128,0, 255,192,128,0,
			255,192,128,0, 255,192,128,0, 255,192,128,0, 255,192,128,0,
			255,192,128,0, 255,192,128,0, 255,192,128,0, 255,192,128,0,
			255,192,128,0, 255,192,128,0, 255,192,128,0, 255,192,128,0,
			255,192,128,0, 255,192,128,0, 255,192,128,0, 255,192,128,0,
			255,192,128,0, 255,192,128,0, 255,192,128,0, 255,192,128,0	};

	// Upload default texture
	defTexObject = Modules::renderer().uploadTextureCube(
		defaultTexture, sizeof(defaultTexture), 0, 2, 2, TextureFormats::RGBX8, false, true );
}


void TextureCubeResource::releaseFunc()
{
	Modules::renderer().unloadTexture( defTexObject, true );
}


TextureCubeResource::TextureCubeResource( const string &name, int flags ) :
	Resource( ResourceTypes::TextureCube, name, flags )
{
	initDefault();
}


TextureCubeResource::~TextureCubeResource()
{
	release();
}


void TextureCubeResource::initDefault()
{
	_width = 0; _height = 0; _format = TextureFormats::RGBX8;
	_texObject = defTexObject;
}


void TextureCubeResource::release()
{
	if( _texObject != 0 && _texObject != defTexObject )
		Modules::renderer().unloadTexture( _texObject, true );

	_texObject = 0;
}


bool TextureCubeResource::raiseError( const string &msg )
{
	// Reset
	release();
	initDefault();

	Modules::log().writeError( "TextureCube resource '%s': %s", _name.c_str(), msg.c_str() );
	
	return false;
}


bool TextureCubeResource::load( const char *data, int size )
{
	if( !Resource::load( data, size ) ) return false;
	
	if( !Modules::config().loadTextures )
		return true;
	
	void *pixels;
	uint32 pixelsSize;
	uint32 mipmaps;
	bool cube;
	if( !ImageLoader::loadImage( data, size, pixels, pixelsSize, _width, _height, _format, mipmaps, cube ) )
		return raiseError( "Invalid image format (" + ImageLoader::getErrorString() + ")" );

	if (!cube)
		return raiseError( "Cube texture is expected" );

	if( !(_flags & ResourceFlags::NoTexMipmaps) && mipmaps == 1)
		mipmaps = 0;	//autogenerate mipmap chain

	_texObject = Modules::renderer().uploadTextureCube(pixels, pixelsSize, mipmaps, _width, _height, _format,
		!(_flags & ResourceFlags::NoTexCompression), !(_flags & ResourceFlags::NoTexFiltering) );

	return true;
}


int TextureCubeResource::getParami( int param )
{
	switch( param )
	{
	case TextureResParams::Width:
		return _width;
	case TextureResParams::Height:
		return _height;
	case TextureResParams::Format:
		return _format;
	default:
		return Resource::getParami( param );
	}
}
